// Copyright 2009, 2010 Christopher Smith
// 
// This file is part of smithers.cards.
// 
// smithers.cards is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
// 
// smithers.cards is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.cards. If not, see <http://www.gnu.org/licenses/>.

package smithers.cards;

import java.util.Arrays;
import java.util.Random;

/**
 * Class to store a stack of {@link Card} objects. Includes methods for
 * creating, sorting, shuffling and dealing cards.
 */
public class CardStack {
    
    private int length;
    private Card[] cardList;
    private Random shuffleGen = new Random();
    
    /**
     * Constructs an empty stack.
     */
    public CardStack() {
        this(0);
    }
    
    /**
     * Creates a stack of a given number of new cards. Creates the first {@code
     * length} cards in the deck (0 and 52 are useful values). Values >52
     * produce multiple decks. <br />
     * eg. 0=empty, 13=all the clubs, 52=whole deck, 104=two decks.
     * 
     * @param length the number of {@code Card}s to create.
     */
    public CardStack(int length) {
        this.length = length;
        cardList = new Card[length];
        for (int i=0; i<length; i++) cardList[i] = new Card(i % 52);
    }
    
    /**
     * Creates a stack from an array of cards.
     *
     * @param cards the {@code Card}s to put into the stack
     */
    public CardStack(Card[] cards) {
        cardList = Arrays.copyOf(cards, cards.length);
        length = cards.length;
    }
    
    /**
     * Returns the number of cards in the list.
     * 
     * @return the number of cards in the list
     */
    public int getLength() {
        return length;
    }
    
    /**
     * Returns the top card in the stack.
     * 
     * @return the card on the top of the stack
     */
    public Card topCard() {
        if (length == 0) throw new RuntimeException("No cards.");
        return cardList[length-1];
    }
    
    /**
     * Returns a specific card from the list.
     * 
     * @param index the index of the card to get
     * @return the card
     */
    public Card getCard(int index) {
        if (index < 0 || index >= length) throw new RuntimeException("Card does not exist.");
        return cardList[index];
    }
    
    /**
     * Returns a {@code String} representation of the entire stack.
     * 
     * @return a {@code String} containing a 2-character representation of each
     *         card in this stack
     */
    public String toString() {
        if (length == 0) return "";
        
        String result = cardList[0].toString();
        for (int i=1; i<length; i++) result += " " + cardList[i].toString();
        return result;
    }
    
    /**
     * Returns an array containing all the cards in this stack. This array will
     * be a copy of this stack so that modifications of either will not affect
     * the other.
     * 
     * @return an array containing a copy of this stack
     */
    public Card[] toArray() {
        return Arrays.copyOf(cardList, length);
    }
    
    // Ensures that the array is at least as large as newSize.
    private void setSize(int newSize) {
        if (newSize > cardList.length) cardList = Arrays.copyOf(cardList, newSize);
    }
    
    /**
     * Adds a card to the top of the stack.
     * 
     * @param newCard the card to add
     */
    public void addCard(Card newCard) {
        setSize(length + 1);
        cardList[length] = newCard;
        length++;
    }
    
    /**
     * Removes a card from the top of the stack.
     */
    public void removeCard() {
        if (length == 0) throw new RuntimeException("No cards to remove.");
        length--;
    }
    
    /**
     * Moves the specified number of cards into the specified stack. If {@code
     * count == 0}, moves all of them. This differs from the method {@code
     * deal(CardStack, count)} in that the cards are kept in the same order.
     * 
     * @param target the stack to deal to
     * @param count  the number of cards to move, or {@code 0} for all of them
     */
    public void move(CardStack target, int count) {
        if (count < 0) throw new RuntimeException("Can't deal negative cards.");
        if (count > length) throw new RuntimeException("Not enough cards to deal.");
        if (count == 0) count = length;
        
        target.setSize(target.length + count);
        for (int i=0; i<count; i++) target.cardList[target.length + i] = cardList[length - count + i];
        target.length += count;
        length -= count;
    }
    
    /**
     * Deals one card into the specified stack.
     * 
     * @param target the stack to deal to
     */
    public void deal(CardStack target) {
        if (length == 0) throw new RuntimeException("No card to deal.");
        target.addCard(cardList[length-1]);
        length--;
    }
    
    /**
     * Deals the specified number of cards into the specified stack. If
     * {@code count == 0}, deals all of them.
     * 
     * @param target the stack to deal to
     * @param count  the number of cards to deal, or {@code 0} for all of them
     */
    public void deal(CardStack target, int count) {
        if (count < 0) throw new RuntimeException("Can't deal negative cards.");
        if (count > length) throw new RuntimeException("Not enough cards to deal.");
        if (count == 0) count = length;
        
        target.setSize(target.length + count);
        for (int i=0; i<count; i++) target.cardList[target.length + i] = cardList[length - i - 1];
        target.length += count;
        length -= count;
    }
    
    /**
     * Deals all of the cards in this stack into the specified stacks.
     * 
     * @param targets the stacks to deal to
     */
    public void deal(CardStack[] targets) {
        deal(targets, 0);
    }
    
    /**
     * Deals the specified number of cards into the specified stacks. If {@code
     * count < 0} then deal {@code -count} cards to each stack, if {@code count
     * > 0} then deal {@code count} cards in total, if {@code count == 0} then
     * deal all the cards.
     * 
     * @param targets the stacks to deal to
     * @param count   the number of cards to deal (as above)
     * @return the number of cards dealt
     */
    public int deal(CardStack[] targets, int count) {
        if (count == 0) count = length;
        else if (count < 0) count *= -targets.length;
        
        if (count > length) throw new RuntimeException("Not enough cards to deal.");
        if (count > 0 && targets.length == 0) throw new RuntimeException("Nowhere to deal to.");
        
        int j = 0;
        for (int i=0; i<count; i++) {
            targets[j].addCard(cardList[length-1]);
            length--;
            j++;
            if (j == targets.length) j = 0;
        }
        return count;
    }
    
    /**
     * Sets the random seed used for shuffling this stack.
     * 
     * @param seed the new seed
     */
    public void setSeed(long seed) {
        shuffleGen.setSeed(seed);
    }
    
    /**
     * Shuffles this stack of cards.
     */
    public void shuffle() {
        int i, j;
        Card temp;
        for (i=length-1; i>0; i--) {
            j = shuffleGen.nextInt(i+1);
            temp = cardList[i];
            cardList[i] = cardList[j];
            cardList[j] = temp;
        }
    }
    
    private int partition(int left, int right, boolean acesHigh) {
        int pivotValue = cardList[right].toInt(acesHigh);
        int j = left;
        Card temp;
        
        for (int i=left; i<right; i++) {
            if (cardList[i].toInt(acesHigh) <= pivotValue) {
                temp = cardList[j];
                cardList[j] = cardList[i];
                cardList[i] = temp;
                j++;
            }
        }
        temp = cardList[j];
        cardList[j] = cardList[right];
        cardList[right] = temp;
        return j;
    }
    
    private void quicksort(int left, int right, boolean acesHigh) {
        if (right > left) {
            int pivotNewIndex = partition(left, right, acesHigh);
            quicksort(left, pivotNewIndex-1, acesHigh);
            quicksort(pivotNewIndex+1, right, acesHigh);
        }
    }
    
    /**
     * Sorts this stack by suit then rank using quicksort with aces low.
     */
    public void sort() {
        quicksort(0, length-1, false);
    }
    
    /**
     * Sorts this stack by suit then rank using quicksort, optionally ranking
     * aces as high or low.
     * 
     * @param acesHigh {@code true} iff aces are to be ranked as high
     */
    public void sort(boolean acesHigh) {
        quicksort(0, length-1, acesHigh);
    }
    
}
