// Copyright 2009 Christopher Smith
// 
// This file is part of smithers.extras.
// 
// smithers.extras is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
// 
// smithers.extras is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.extras. If not, see <http://www.gnu.org/licenses/>.

package smithers.extras;

public class Cellular2D {
    
    public static class Ruleset {
        
        private int bornFlags, survivesFlags;
        
        public Ruleset(int bornFlags, int survivesFlags) {
            this.bornFlags = bornFlags;
            this.survivesFlags = survivesFlags;
        }
        
        public Ruleset(int[] born, int[] survives) {
            bornFlags = 0;
            survivesFlags = 0;
            for (int b : born) bornFlags |= 1<<b;
            for (int d : survives) survivesFlags |= 1<<d;
        }
        
        public Ruleset(Iterable<Integer> born, Iterable<Integer> survives) {
            bornFlags = 0;
            survivesFlags = 0;
            for (int b : born) bornFlags |= 1<<b;
            for (int d : survives) survivesFlags |= 1<<d;
        }
        
        public boolean born(int neighbours) {
            return ((bornFlags>>neighbours)&1) != 0;
        }
        
        public boolean survives(int neighbours) {
            return ((survivesFlags>>neighbours)&1) != 0;
        }
        
    }
    
    public static final Ruleset GAME_OF_LIFE = new Ruleset(0x08, 0x0c);
    public static final Ruleset HIGHLIFE = new Ruleset(0x48, 0x0c);
    public static final Ruleset DAY_NIGHT = new Ruleset(0x1c8, 0x1d8);
    
    private Ruleset rules;
    private int width;
    private int height;
    private boolean[][] grid;
    
    public Cellular2D(Ruleset rules, int width, int height) {
        this.rules = rules;
        this.width = width;
        this.height = height;
        this.grid = new boolean[width][height];
    }
    
    public Cellular2D(Ruleset rules, boolean[][] grid) {
        this.rules = rules;
        this.width = grid.length;
        this.height = grid[0].length;
        this.grid = new boolean[width][height];
        for (int j=0; j<height; j++)
            for (int i=0; i<width; i++)
                this.grid[i][j] = grid[i][j];
    }
    
    public Ruleset getRules() {
        return rules;
    }
    
    public void randomise(double p) {
        for (int j=0; j<height; j++)
            for (int i=0; i<width; i++)
                grid[i][j] = Math.random()<p;
    }
    
    public void clear() {
        for (int j=0; j<height; j++)
            for (int i=0; i<width; i++)
                grid[i][j] = false;
    }
    
    public void step() {
        boolean[][] newGrid = new boolean[width][height];
        for (int j=0; j<height; j++)
            for (int i=0; i<width; i++)
                if (grid[i][j])
                    newGrid[i][j] = rules.survives(countNb(i,j));
                else
                    newGrid[i][j] = rules.born(countNb(i,j));
        grid = newGrid;
    }
    
    private int countNb(int x, int y) {
        int result = 0;
        for (int j=y-1; j<y+2; j++)
            for (int i=x-1; i<x+2; i++)
                if ((i!=x || j!=y) && grid[i>=width?0:i<0?width-1:i][j>=height?0:j<0?height-1:j])
                    result++;
        return result;
    }
    
    public String gridString() {
        StringBuilder result = new StringBuilder((width+3)*(height+2));
        result.append('+');
        for (int i=0; i<width; i++)
            result.append('-');
        result.append("+\n");
        for (int j=0; j<height; j++) {
            result.append('|');
            for (int i=0; i<width; i++)
                result.append(grid[i][j]?'#':' ');
            result.append("|\n");
        }
        result.append('+');
        for (int i=0; i<width; i++)
            result.append('-');
        result.append("+\n");
        return result.toString();
    }
    
    public static void main(String[] args) throws java.io.IOException {
        if (args.length > 0 && (args[0].equals("-h") || args[0].equals("--help"))) {
            System.out.println("smithers.extras.Cellular2D -- simulates 2-dimensional cellular automata.");
            return;
        }
        Cellular2D c = new Cellular2D(GAME_OF_LIFE, 70, 30);
        c.randomise(0.2);
        System.out.print(c.gridString());
        while (System.in.read()!=-1) {
            c.step();
            System.out.println(c.gridString());
        }
    }
    
}
