// Copyright 2010 Christopher Smith
// 
// This file is part of smithers.games.
// 
// smithers.games is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
// 
// smithers.games is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.games. If not, see <http://www.gnu.org/licenses/>.

package smithers.games;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.util.*;

/**
 * A light-cycles game commonly associated with the film 'Tron'.
 */
public class JTron extends Canvas implements ComponentListener, KeyListener, MouseListener, Runnable {
    
    static enum GameState {
        STOPPED, READY, RUNNING, ENDED;
    }
    
    static enum Direction {
        NORTH, EAST, SOUTH, WEST;
        Direction opposite() {
            switch (this) {
                case NORTH: return SOUTH;
                case EAST: return WEST;
                case SOUTH: return NORTH;
                default: return EAST;
            }
        }
        int gridMask() {
            switch (this) {
                case NORTH: return GRIDMASK_N;
                case EAST: return GRIDMASK_E;
                case SOUTH: return GRIDMASK_S;
                default: return GRIDMASK_W;
            }
        }
    }
    
    static class Player {
        Point position;
        int score = 0;
        Direction direction;
        Direction lastDir;
        boolean isReady;
        boolean crashed;
        boolean accelerated;
    }
    
    static final Color COL_BG = Color.BLACK;
    static final Color[] COL_PLAYER = {Color.RED, Color.BLUE, Color.GREEN, Color.YELLOW};
    static final int GRIDMASK_N = 1, GRIDMASK_E = 2, GRIDMASK_S = 4, GRIDMASK_W = 8;
    static final int[] GRIDMASK_PLAYER = {16, 32, 48, 64};
    static final int[] KEY_N = {KeyEvent.VK_W, KeyEvent.VK_UP, KeyEvent.VK_I, KeyEvent.VK_T};
    static final int[] KEY_E = {KeyEvent.VK_D, KeyEvent.VK_RIGHT, KeyEvent.VK_L, KeyEvent.VK_H};
    static final int[] KEY_S = {KeyEvent.VK_S, KeyEvent.VK_DOWN, KeyEvent.VK_K, KeyEvent.VK_G};
    static final int[] KEY_W = {KeyEvent.VK_A, KeyEvent.VK_LEFT, KeyEvent.VK_J, KeyEvent.VK_F};
    static final int[] KEY_A = {KeyEvent.VK_Z, KeyEvent.VK_CONTROL, KeyEvent.VK_V, KeyEvent.VK_M};
    
    final int playerCount;
    final Player[] player;
    final int frameTime;
    final int gridDrawSize;
    int gridHeight, gridWidth;
    int[][] grid;
    GameState gameState;
    Thread gameThread;
    BufferedImage screenImage;
    Graphics2D screenGraphics;
    boolean needToRefresh;
    
    Panel statusBar;
    Label scoreLabel[];
    
    public JTron(int playerCount, int frameTime, int gridDrawSize) {
        this.playerCount = playerCount;
        this.player = new Player[playerCount];
        this.frameTime = frameTime;
        this.gridDrawSize = gridDrawSize;
        addKeyListener(this);
        addMouseListener(this);
        gameThread = new Thread(this);
        gameThread.setDaemon(true);
        gameState = GameState.STOPPED;
        gameThread.start();
        
        statusBar = new Panel();
        statusBar.setLayout(new GridLayout(1, playerCount));
        scoreLabel = new Label[playerCount];
        for (int i=0; i<playerCount; i++)
            statusBar.add(scoreLabel[i] = new Label("Player "+(i+1)+": 0"));
    }
    
    public void reset() {
        for (int i=0; i<playerCount; i++)
            player[i] = new Player();
        newGame();
    }
    
    public void newGame() {
        gameState = GameState.STOPPED;
        
        Dimension d = getSize();
        gridHeight = d.height / gridDrawSize;
        gridWidth = d.width / gridDrawSize;
        grid = new int[gridHeight][gridWidth];
        screenImage = new BufferedImage(gridWidth * gridDrawSize, gridHeight * gridDrawSize, BufferedImage.TYPE_INT_RGB);
        screenGraphics = screenImage.createGraphics();
        
        int i, j, k;
        int[] shuffle = new int[playerCount];
        for (i=0; i<playerCount; i++) shuffle[i] = i;
        Random r = new Random();
        for (i=0; i<playerCount; i++) {
            j = r.nextInt(playerCount - i) + i;
            k = shuffle[i];
            shuffle[i] = shuffle[j];
            shuffle[j] = k;
        }
        
        if (playerCount == 2) {
            player[shuffle[0]].position = new Point(gridWidth / 3, gridHeight / 2);
            player[shuffle[1]].position = new Point(gridWidth - 1 - gridWidth / 3, gridHeight / 2);
        } else if (playerCount == 3) {
            player[shuffle[0]].position = new Point(1 * gridWidth / 4, 2 * gridHeight / 3);
            player[shuffle[1]].position = new Point(2 * gridWidth / 4, 1 * gridHeight / 3);
            player[shuffle[2]].position = new Point(3 * gridWidth / 4, 2 * gridHeight / 3);
        } else if (playerCount == 4) {
            int x1 = gridWidth / 3, x2 = gridWidth - 1 - x1;
            int y1 = gridHeight / 3, y2 = gridHeight - 1 - y1;
            player[shuffle[0]].position = new Point(x1, y1);
            player[shuffle[1]].position = new Point(x1, y2);
            player[shuffle[2]].position = new Point(x2, y1);
            player[shuffle[3]].position = new Point(x2, y2);
        }
        for (i=0; i<playerCount; i++) {
            player[i].isReady = false;
            player[i].crashed = false;
            grid[player[i].position.y][player[i].position.x] = GRIDMASK_PLAYER[i];
            gridPaintCell(player[i].position);
        }
        
        repaint();
        gameState = GameState.READY;
    }
    
    void gridAddConnection(Point p, Direction d) {
        grid[p.y][p.x] |= d.gridMask();
    }
    
    Color lighter(Color col) {
        int r, g, b;
        r = col.getRed() + 128;
        if (r > 255) r = 255;
        g = col.getGreen() + 128;
        if (g > 255) g = 255;
        b = col.getBlue() + 128;
        if (b > 255) b = 255;
        return new Color(r, g, b);
    }
    
    Color darker(Color col) {
        int r, g, b;
        r = col.getRed() - 128;
        if (r < 0) r = 0;
        g = col.getGreen() - 128;
        if (g < 0) g = 0;
        b = col.getBlue() - 128;
        if (b < 0) b = 0;
        return new Color(r, g, b);
    }
    
    void gridPaintCell(Point p) {
        int gridValue = grid[p.y][p.x];
        int x = p.x * gridDrawSize;
        int y = p.y * gridDrawSize;
        int playerIndex = -1;
        Color col = COL_BG;
        
        for (int i=0; i<playerCount; i++)
            if ((gridValue & GRIDMASK_PLAYER[i]) == GRIDMASK_PLAYER[i]) playerIndex = i;
        if (playerIndex >= 0)
            col = COL_PLAYER[playerIndex];
        
        screenGraphics.setColor(col);
        screenGraphics.fillRect(x, y, gridDrawSize, gridDrawSize);
        if (playerIndex != -1) {
            screenGraphics.setColor(lighter(col));
            if ((gridValue & GRIDMASK_N) == 0) screenGraphics.drawLine(x, y, x + gridDrawSize - 1, y);
            if ((gridValue & GRIDMASK_W) == 0) screenGraphics.drawLine(x, y, x, y + gridDrawSize - 1);
            screenGraphics.setColor(darker(col));
            if ((gridValue & GRIDMASK_S) == 0) screenGraphics.drawLine(x, y + gridDrawSize - 1, x + gridDrawSize - 1, y + gridDrawSize - 1);
            if ((gridValue & GRIDMASK_E) == 0) screenGraphics.drawLine(x + gridDrawSize - 1, y, x + gridDrawSize - 1, y + gridDrawSize - 1);
        }
        repaint(x, y, gridDrawSize, gridDrawSize);
    }
    
    void step() {
        Point[] moveTo = new Point[playerCount];
        boolean[] crashing = new boolean[playerCount];
        int crashCount = 0, aliveCount = 0;
        int i, j;
        
        for (i=0; i<playerCount; i++) {
            if (player[i].crashed) continue;
            crashing[i] = false;
            moveTo[i] = new Point(player[i].position);
            switch (player[i].direction) {
                case NORTH: moveTo[i].y--; break;
                case EAST: moveTo[i].x++; break;
                case SOUTH: moveTo[i].y++; break;
                case WEST: moveTo[i].x--; break;
            }
            player[i].lastDir = player[i].direction;
            if (moveTo[i].x < 0 || moveTo[i].x >= gridWidth || moveTo[i].y < 0 || moveTo[i].y >= gridHeight || grid[moveTo[i].y][moveTo[i].x] != 0) {
                crashing[i] = true;
                crashCount++;
            } else {
                for (j=0; j<i; j++)
                    if (moveTo[i].equals(moveTo[j])) {
                        crashCount += crashing[j] ? 1 : 2;
                        crashing[i] = crashing[j] = true;
                        break;
                    }
            }
        }
        
        for (i=0; i<playerCount; i++) {
            if (player[i].crashed) continue;
            if (crashing[i]) {
                player[i].score += crashCount - 1;
                System.out.println("Player " + (i+1) + " has crashed!");
                player[i].crashed = true;
            } else {
                player[i].score += crashCount;
                gridAddConnection(player[i].position, player[i].direction);
                gridPaintCell(player[i].position);
                player[i].position.setLocation(moveTo[i]);
                grid[player[i].position.y][player[i].position.x] = GRIDMASK_PLAYER[i];
                gridAddConnection(player[i].position, player[i].direction.opposite());
                gridPaintCell(player[i].position);
            }
        }
        
        crashCount = 0;
        for (i=0; i<playerCount; i++) {
            if (player[i].crashed || !player[i].accelerated) continue;
            crashing[i] = false;
            moveTo[i] = new Point(player[i].position);
            switch (player[i].direction) {
                case NORTH: moveTo[i].y--; break;
                case EAST: moveTo[i].x++; break;
                case SOUTH: moveTo[i].y++; break;
                case WEST: moveTo[i].x--; break;
            }
            player[i].lastDir = player[i].direction;
            if (moveTo[i].x < 0 || moveTo[i].x >= gridWidth || moveTo[i].y < 0 || moveTo[i].y >= gridHeight || grid[moveTo[i].y][moveTo[i].x] != 0) {
                crashing[i] = true;
                crashCount++;
            } else {
                for (j=0; j<i; j++)
                    if (moveTo[i].equals(moveTo[j])) {
                        crashCount += crashing[j] ? 1 : 2;
                        crashing[i] = crashing[j] = true;
                        break;
                    }
            }
        }
        
        for (i=0; i<playerCount; i++) {
            if (!player[i].crashed && player[i].accelerated) {
                if (crashing[i]) {
                    player[i].score += crashCount - 1;
                    player[i].crashed = true;
                } else {
                    player[i].score += crashCount;
                    gridAddConnection(player[i].position, player[i].direction);
                    gridPaintCell(player[i].position);
                    player[i].position.setLocation(moveTo[i]);
                    grid[player[i].position.y][player[i].position.x] = GRIDMASK_PLAYER[i];
                    gridAddConnection(player[i].position, player[i].direction.opposite());
                    gridPaintCell(player[i].position);
                }
            }
            if (!player[i].crashed) aliveCount++;
            scoreLabel[i].setText("Player " + (i+1) + ": " + player[i].score);
        }
        
        if (aliveCount <= 1) gameState = GameState.ENDED;
    }
    
    public void update(Graphics g) {
        if (needToRefresh) {
            g.setColor(getBackground());
            g.fillRect(0, 0, getWidth(), getHeight());
            needToRefresh = false;
        }
        ((Graphics2D)g).drawImage(screenImage, null, 0, 0);
    }
    
    public void run() {
        long tm = System.currentTimeMillis();
        while (Thread.currentThread() == gameThread) {
            if (gameState == GameState.RUNNING) {
                step();
            }
            if (gameState == GameState.ENDED) tm += 1000;
            try {
                tm += frameTime;
                long ctm = System.currentTimeMillis();
                if (tm < ctm) tm = ctm;
                Thread.sleep(tm - ctm);
            } catch (InterruptedException e) {
                break;
            }
            if (gameState == GameState.ENDED) gameState = GameState.STOPPED;
        }
    }
    
    public void keyPressed(KeyEvent e) {
        boolean canStart = gameState == GameState.READY;
        for (int i=0; i<playerCount; i++) {
            if (e.getKeyCode() == KEY_N[i]) {
                if (gameState != GameState.RUNNING || player[i].lastDir != Direction.NORTH.opposite())
                    player[i].direction = Direction.NORTH;
                if (gameState == GameState.READY) player[i].isReady = true;
                
            } else if (e.getKeyCode() == KEY_E[i]) {
                if (gameState != GameState.RUNNING || player[i].lastDir != Direction.EAST.opposite())
                    player[i].direction = Direction.EAST;
                if (gameState == GameState.READY) player[i].isReady = true;
                
            } else if (e.getKeyCode() == KEY_S[i]) {
                if (gameState != GameState.RUNNING || player[i].lastDir != Direction.SOUTH.opposite())
                    player[i].direction = Direction.SOUTH;
                if (gameState == GameState.READY) player[i].isReady = true;
                
            } else if (e.getKeyCode() == KEY_W[i]) {
                if (gameState != GameState.RUNNING || player[i].lastDir != Direction.WEST.opposite())
                    player[i].direction = Direction.WEST;
                if (gameState == GameState.READY) player[i].isReady = true;
                
            } else if (e.getKeyCode() == KEY_A[i]) {
                player[i].accelerated = true;
            }
            canStart &= player[i].isReady;
        }
        if (canStart) gameState = GameState.RUNNING;
        if (gameState == GameState.STOPPED) newGame();
    }
    
    public void keyReleased(KeyEvent e) {
        for (int i=0; i<playerCount; i++)
            if (e.getKeyCode() == KEY_A[i])
                player[i].accelerated = false;
    }
    
    
    public void mouseClicked(MouseEvent e) {
        newGame();
    }
    
    public void componentResized(ComponentEvent e) {
        needToRefresh = true;
    }
    
    public void componentShown(ComponentEvent e) {
        needToRefresh = true;
    }
    
    public void componentHidden(ComponentEvent e) {}
    public void componentMoved(ComponentEvent e) {}
    public void keyTyped(KeyEvent e) {}
    public void mouseEntered(MouseEvent e) {}
    public void mouseExited(MouseEvent e) {}
    public void mousePressed(MouseEvent e) {}
    public void mouseReleased(MouseEvent e) {}
    
    public static void main(String[] args) {
        int playerCount = 2;
        int frameTime = 100;
        int gridDrawSize = 10;
        int temp;
        for (int i=0; i<args.length; i++) {
            if (args[i].equals("-h") || args[i].equals("--help")) {
                System.out.println("smithers.games.JTron -- light cycles game.");
                System.out.println("Usage: java smithers.games.JTron [options...]");
                System.out.println("Options:");
                System.out.println("  -p n  Set the number of players to n.");
                System.out.println("  -s n  Set the grid size to n.");
                System.out.println("  -t n  Set the speed to 1 frame per n milliseconds.");
                System.out.println("  -h");
                System.out.println("  --help");
                System.out.println("        Display this help message and exit.");
                System.out.println("");
                System.out.println("In-game controls:");
                System.out.println("  Player 1 (Red): W/A/S/D");
                System.out.println("  Player 2 (Blue): Cursor Up/Left/Down/Right");
                System.out.println("  Player 3 (Green): I/J/K/L");
                System.out.println("  Player 4 (Yellow): T/F/G/H");
                System.out.println("Have Fun!");
                return;
            } else if (args[i].equals("-p")) {
                temp = Integer.parseInt(args[++i]);
                if (temp >= 2 && temp <= 4) playerCount = temp;
            } else if (args[i].equals("-s")) {
                gridDrawSize = Integer.parseInt(args[++i]);
            } else if (args[i].equals("-t")) {
                frameTime = Integer.parseInt(args[++i]);
            }
        }
        final JTron game = new JTron(playerCount, frameTime, gridDrawSize);
        
        EventQueue.invokeLater(new Runnable() {
            public void run() {
                Frame frame = new Frame("JTron by Smithers");
                frame.addWindowListener(new WindowAdapter() {
                    public void windowClosing(WindowEvent we) {
                        we.getWindow().dispose();
                    }
                });
                frame.setSize(800, 600);
                frame.setLayout(new BorderLayout());
                frame.add(game);
                frame.add(game.statusBar, BorderLayout.SOUTH);
                game.requestFocusInWindow();
                frame.setVisible(true);
                game.reset();
                game.needToRefresh = true;
            }
        });
    }
    
}
