// Copyright 2010 Christopher Smith
// 
// This file is part of smithers.logicpuzzles.
// 
// smithers.logicpuzzles is free software: you can redistribute it and/or
// modify it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
// 
// smithers.logicpuzzles is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.logicpuzzles. If not, see <http://www.gnu.org/licenses/>.

package smithers.logicpuzzles;

import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import javax.imageio.ImageIO;

/**
 * Creates hanjie puzzles from image files.
 */
public class HanjieBuilder {
    
    private HanjieBuilder() {}
    
    /**
     * Creates a new hanjie puzzle based on an image file. This should be
     * called with a single argument, the name of the file. The puzzle will
     * then be written to {@code System.out}. This will only handle images with
     * up to 26 colours plus blank. Each pixel of the image is taken to be a
     * cell in the puzzle. Blank must be pure white, and colours are considered
     * equal iff they are exactly equal, so lossy image formats (such as JPEG)
     * are not practical.
     * 
     * @param args the arguments, of which the first should be the name of an
     *             image file
     */
    public static void main(String[] args) throws IOException {
        if (args.length == 0 || args[0].equals("-h") || args[0].equals("--help")) {
            System.out.println("smithers.logicpuzzles.HanjieBuilder -- Creates hanjie puzzles from images.");
            System.out.println("Usage:");
            System.out.println("        java smithers.logicpuzzles.HanjieBuilder -h|--help");
            System.out.println("        java smithers.logicpuzzles.HanjieBuilder image");
            System.out.println("");
            System.out.println("If -h or --help is given, display this help and exit.");
            System.out.println("Otherwise image must be the path to an image from which to create a hanjie puzzle. The puzzle will be printed to standard output.");
            System.out.println("The image should be of the same size as the puzzle to be produced, and can contain up to 26 colours, plus white which is used as the blank colour.");
            return;
        }
        
        BufferedImage image = ImageIO.read(new File(args[0]));
        int height = image.getHeight();
        int width = image.getWidth();
        int[] rgbArray = image.getRGB(0, 0, width, height, null, 0, width);
        int cols = 1;
        int[] colArray = new int[27];
        colArray[0] = 0xffffffff;
        StringBuilder output = new StringBuilder();
        int ccol, length;
        int i, j;
        
        for (i=0; i<height*width; i++) {
            for (j=0; j<cols; j++)
                if (rgbArray[i] == colArray[j]) break;
            if (j==cols) {
                if (cols==27) throw new RuntimeException("Too many colours - ensure equal colours are exactly equal.");
                colArray[cols] = rgbArray[i];
                cols++;
            }
        }
        
        if (cols > 2) {
            System.out.println('C');
            output.append("ffffff");
            for (i=1; i<cols; i++) {
                output.append(String.format(" %c %06x", 'a'+i-1, colArray[i] & 0xffffff));
            }
            System.out.println(output);
            output.delete(0, output.length());
        }
        System.out.println(height + " " + width);
        
        for (i=0; i<height; i++) {
            ccol = 0;
            length = 0;
            for (j=0; j<width; j++) {
                if (rgbArray[i*width+j] == colArray[ccol]) {
                    length++;
                } else {
                    if (ccol != 0) {
                        if (output.length() > 0) output.append(' ');
                        if (cols > 2) output.append((char)('a'+ccol-1));
                        output.append(length);
                    }
                    for (ccol=0; colArray[ccol] != rgbArray[i*width+j]; ccol++);
                    length = 1;
                }
            }
            if (ccol != 0) {
                if (output.length() > 0) output.append(' ');
                if (cols > 2) output.append((char)('a'+ccol-1));
                output.append(length);
            }
            System.out.println(output);
            output.delete(0, output.length());
        }
        for (j=0; j<width; j++) {
            ccol = 0;
            length = 0;
            for (i=0; i<height; i++) {
                if (rgbArray[i*width+j] == colArray[ccol]) {
                    length++;
                } else {
                    if (ccol != 0) {
                        if (output.length() > 0) output.append(' ');
                        if (cols > 2) output.append((char)('a'+ccol-1));
                        output.append(length);
                    }
                    for (ccol=0; colArray[ccol] != rgbArray[i*width+j]; ccol++);
                    length = 1;
                }
            }
            if (ccol != 0) {
                if (output.length() > 0) output.append(' ');
                if (cols > 2) output.append((char)('a'+ccol-1));
                output.append(length);
            }
            System.out.println(output);
            output.delete(0, output.length());
        }
    }
    
}
