// Copyright 2010 Christopher Smith
// 
// This file is part of smithers.logicpuzzles.
// 
// smithers.logicpuzzles is free software: you can redistribute it and/or
// modify it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
// 
// smithers.logicpuzzles is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.logicpuzzles. If not, see <http://www.gnu.org/licenses/>.

package smithers.logicpuzzles;

import smithers.util.*;

/**
 * Solver for slitherlink puzzles.
 */
public class SlitherlinkSolver extends AbstractPuzzleSolver<Slitherlink, SlitherlinkState> {
    
    private static final int MASK_ALL = (1<<16)-1;
    
    private static final int MASK_ZERO = (1<<0);
    private static final int MASK_ONE = (1<<1)|(1<<2)|(1<<4)|(1<<8);
    private static final int MASK_TWO = (1<<3)|(1<<5)|(1<<6)|(1<<9)|(1<<10)|(1<<12);
    private static final int MASK_THREE = (1<<7)|(1<<11)|(1<<13)|(1<<14);
    private static final int MASK_FOUR = (1<<15);
    
    private static final int MASK_RIGHT = (1<<1)|(1<<3)|(1<<5)|(1<<7)|(1<<9)|(1<<11)|(1<<13)|(1<<15);
    private static final int MASK_UP = (1<<2)|(1<<3)|(1<<6)|(1<<7)|(1<<10)|(1<<11)|(1<<14)|(1<<15);
    private static final int MASK_LEFT = (1<<4)|(1<<5)|(1<<6)|(1<<7)|(1<<12)|(1<<13)|(1<<14)|(1<<15);
    private static final int MASK_DOWN = (1<<8)|(1<<9)|(1<<10)|(1<<11)|(1<<12)|(1<<13)|(1<<14)|(1<<15);
    
    private static final int MASK_UP_AND_LEFT = MASK_UP & MASK_LEFT;
    private static final int MASK_UP_XOR_LEFT = ((~MASK_UP & MASK_LEFT) | (MASK_UP & ~MASK_LEFT));
    private static final int MASK_UP_NOR_LEFT = ~MASK_UP & ~MASK_LEFT;
    private static final int MASK_UP_AND_RIGHT = MASK_UP & MASK_RIGHT;
    private static final int MASK_UP_XOR_RIGHT = ((~MASK_UP & MASK_RIGHT) | (MASK_UP & ~MASK_RIGHT));
    private static final int MASK_UP_NOR_RIGHT = ~MASK_UP & ~MASK_RIGHT;
    private static final int MASK_DOWN_AND_LEFT = MASK_DOWN & MASK_LEFT;
    private static final int MASK_DOWN_XOR_LEFT = ((~MASK_DOWN & MASK_LEFT) | (MASK_DOWN & ~MASK_LEFT));
    private static final int MASK_DOWN_NOR_LEFT = ~MASK_DOWN & ~MASK_LEFT;
    private static final int MASK_DOWN_AND_RIGHT = MASK_DOWN & MASK_RIGHT;
    private static final int MASK_DOWN_XOR_RIGHT = ((~MASK_DOWN & MASK_RIGHT) | (MASK_DOWN & ~MASK_RIGHT));
    private static final int MASK_DOWN_NOR_RIGHT = ~MASK_DOWN & ~MASK_RIGHT;
    
    private int height, width;
    private int cellI, cellJ;
    // passType:
    // -1 => special
    // 0 => normal pass NW->SE
    // 1 => loopcheck horizontal
    // 2 => normal pass SE->NW
    // 3 => loopcheck vertical
    private int passType;
    private int[][] data; // Indexes into data are offset by (-1,-1)
    private UnionFindNode[][] vertexLinks;
    private AlternatingUnionFindNode[][] cellLinks;
    
    /**
     * Creates a new solver for the specified puzzle.
     * 
     * @param puzzle the puzzle to solve
     */
    public SlitherlinkSolver(Slitherlink puzzle) {
        super(puzzle);
        height = puzzle.height;
        width = puzzle.width;
        start();
    }
    
    /**
     * Creates a new solver to work from the specified state.
     * 
     * @param state the puzzle state to work from
     */
    public SlitherlinkSolver(SlitherlinkState state) {
        super(state);
        height = puzzle.height;
        width = puzzle.width;
        start();
    }
    
    public SlitherlinkState makeState(Slitherlink puzzle) {
        return new SlitherlinkState(puzzle);
    }
    
    public void start() {
        int i, j;
        
        data = new int[height+2][width+2];
        for (i=-1; i<height+1; i++) for (j=-1; j<width+1; j++) {
            if ((i == -1 || i == height) && (j == -1 || j == width)) {
                data[i+1][j+1] = MASK_ZERO;
            } else if (i == -1) {
                switch (state.getHorizontalState(i+1, j)) {
                    case 0: data[i+1][j+1] = MASK_ZERO; break;
                    case 1: data[i+1][j+1] = MASK_ONE & MASK_DOWN; break;
                    default: data[i+1][j+1] = MASK_ZERO | (MASK_ONE & MASK_DOWN);
                }
            } else if (i == height) {
                switch (state.getHorizontalState(i, j)) {
                    case 0: data[i+1][j+1] = MASK_ZERO; break;
                    case 1: data[i+1][j+1] = MASK_ONE & MASK_UP; break;
                    default: data[i+1][j+1] = MASK_ZERO | (MASK_ONE & MASK_UP);
                }
            } else if (j == -1) {
                switch (state.getVerticalState(i, j+1)) {
                    case 0: data[i+1][j+1] = MASK_ZERO; break;
                    case 1: data[i+1][j+1] = MASK_ONE & MASK_RIGHT; break;
                    default: data[i+1][j+1] = MASK_ZERO | (MASK_ONE & MASK_RIGHT);
                }
            } else if (j == width) {
                switch (state.getVerticalState(i, j)) {
                    case 0: data[i+1][j+1] = MASK_ZERO; break;
                    case 1: data[i+1][j+1] = MASK_ONE & MASK_LEFT; break;
                    default: data[i+1][j+1] = MASK_ZERO | (MASK_ONE & MASK_LEFT);
                }
            } else {
                switch (puzzle.getClue(i, j)) {
                    case 0: data[i+1][j+1] = MASK_ZERO; break;
                    case 1: data[i+1][j+1] = MASK_ONE; break;
                    case 2: data[i+1][j+1] = MASK_TWO; break;
                    case 3: data[i+1][j+1] = MASK_THREE; break;
                    case 4: data[i+1][j+1] = MASK_FOUR; break;
                    default: data[i+1][j+1] = MASK_ALL;
                }
                switch (state.getVerticalState(i, j+1)) {
                    case 0: data[i+1][j+1] &= ~MASK_RIGHT; break;
                    case 1: data[i+1][j+1] &= MASK_RIGHT; break;
                }
                switch (state.getHorizontalState(i, j)) {
                    case 0: data[i+1][j+1] &= ~MASK_UP; break;
                    case 1: data[i+1][j+1] &= MASK_UP; break;
                }
                switch (state.getVerticalState(i, j)) {
                    case 0: data[i+1][j+1] &= ~MASK_LEFT; break;
                    case 1: data[i+1][j+1] &= MASK_LEFT; break;
                }
                switch (state.getHorizontalState(i+1, j)) {
                    case 0: data[i+1][j+1] &= ~MASK_DOWN; break;
                    case 1: data[i+1][j+1] &= MASK_DOWN; break;
                }
            }
        }
        
        vertexLinks = new UnionFindNode[height+1][width+1];
        for (i=0; i<height+1; i++) for (j=0; j<width+1; j++)
            vertexLinks[i][j] = new UnionFindNode();
        for (i=0; i<height+1; i++) for (j=0; j<width; j++)
            if (state.getHorizontalState(i, j) == 1)
                vertexLinks[i][j].union(vertexLinks[i][j+1]);
        for (i=0; i<height; i++) for (j=0; j<width+1; j++)
            if (state.getVerticalState(i, j) == 1)
                vertexLinks[i][j+1].union(vertexLinks[i][j]);
        
        cellLinks = new AlternatingUnionFindNode[height][width];
        for (i=0; i<height; i++) for (j=0; j<width; j++)
            cellLinks[i][j] = new AlternatingUnionFindNode();
        for (i=1; i<height; i++) for (j=0; j<width; j++)
            if (state.getHorizontalState(i, j) != -1)
                cellLinks[i-1][j].union(cellLinks[i][j], state.getHorizontalState(i, j) == 0);
        for (i=0; i<height; i++) for (j=1; j<width; j++)
            if (state.getVerticalState(i, j) != -1)
                cellLinks[i][j-1].union(cellLinks[i][j], state.getVerticalState(i, j) == 0);
        
        cellI = cellJ = 0;
        passType = -1;
    }
    
    public int solveStep() {
        int result = 0;
        
        if (passType == 0 || passType == 2) {
            int value = data[cellI+1][cellJ+1];
            int corner;
            
            if (cellI > -1) {
                if (cellJ > -1) {
                    corner = data[cellI][cellJ];
                    if ((corner & MASK_DOWN_XOR_RIGHT) == 0)
                        value &= ~MASK_UP_XOR_LEFT;
                    if ((corner & ~MASK_DOWN_XOR_RIGHT) == 0)
                        value &= MASK_UP_XOR_LEFT;
                    if ((corner & MASK_DOWN_NOR_RIGHT) == 0)
                        value &= ~MASK_UP_AND_LEFT;
                }
                if (cellJ < width) {
                    corner = data[cellI][cellJ+2];
                    if ((corner & MASK_DOWN_XOR_LEFT) == 0)
                        value &= ~MASK_UP_XOR_RIGHT;
                    if ((corner & ~MASK_DOWN_XOR_LEFT) == 0)
                        value &= MASK_UP_XOR_RIGHT;
                    if ((corner & MASK_DOWN_NOR_LEFT) == 0)
                        value &= ~MASK_UP_AND_RIGHT;
                }
            }
            if (cellI < height) {
                if (cellJ > -1) {
                    corner = data[cellI+2][cellJ];
                    if ((corner & MASK_UP_XOR_RIGHT) == 0)
                        value &= ~MASK_DOWN_XOR_LEFT;
                    if ((corner & ~MASK_UP_XOR_RIGHT) == 0)
                        value &= MASK_DOWN_XOR_LEFT;
                    if ((corner & MASK_UP_NOR_RIGHT) == 0)
                        value &= ~MASK_DOWN_AND_LEFT;
                }
                if (cellJ < width) {
                    corner = data[cellI+2][cellJ+2];
                    if ((corner & MASK_UP_XOR_LEFT) == 0)
                        value &= ~MASK_DOWN_XOR_RIGHT;
                    if ((corner & ~MASK_UP_XOR_LEFT) == 0)
                        value &= MASK_DOWN_XOR_RIGHT;
                    if ((corner & MASK_UP_NOR_LEFT) == 0)
                        value &= ~MASK_DOWN_AND_RIGHT;
                }
            }
            
            if (value != data[cellI+1][cellJ+1]) {
                data[cellI+1][cellJ+1] = value;
                result = 1;
            }
            
            if (cellI > -1 && cellI < height) {
                if (cellJ > -1 && state.getVerticalState(cellI, cellJ) == -1) {
                    if ((value & MASK_LEFT) == 0) {
                        clearVertical(cellI, cellJ);
                        result = 2;
                    } else if ((value & ~MASK_LEFT) == 0) {
                        fillVertical(cellI, cellJ);
                        result = 2;
                    }
                }
                if (cellJ < width && state.getVerticalState(cellI, cellJ+1) == -1) {
                    if ((value & MASK_RIGHT) == 0) {
                        clearVertical(cellI, cellJ+1);
                        result = 2;
                    } else if ((value & ~MASK_RIGHT) == 0) {
                        fillVertical(cellI, cellJ+1);
                        result = 2;
                    }
                }
            }
            if (cellJ > -1 && cellJ < width) {
                if (cellI > -1 && state.getHorizontalState(cellI, cellJ) == -1) {
                    if ((value & MASK_UP) == 0) {
                        clearHorizontal(cellI, cellJ);
                        result = 2;
                    } else if ((value & ~MASK_UP) == 0) {
                        fillHorizontal(cellI, cellJ);
                        result = 2;
                    }
                }
                if (cellI < height && state.getHorizontalState(cellI+1, cellJ) == -1) {
                    if ((value & MASK_DOWN) == 0) {
                        clearHorizontal(cellI+1, cellJ);
                        result = 2;
                    } else if ((value & ~MASK_DOWN) == 0) {
                        fillHorizontal(cellI+1, cellJ);
                        result = 2;
                    }
                }
            }
            
        } else if (passType == 1) {
            if (state.getHorizontalState(cellI, cellJ) == -1) {
                if (vertexLinks[cellI][cellJ].sameSet(vertexLinks[cellI][cellJ+1])) {
                    clearHorizontal(cellI, cellJ);
                    result = 2;
                } else if (cellI > 0 && cellI < height && cellLinks[cellI-1][cellJ].sameSet(cellLinks[cellI][cellJ])) {
                    if (cellLinks[cellI-1][cellJ].samePart(cellLinks[cellI][cellJ]))
                        clearHorizontal(cellI, cellJ);
                    else
                        fillHorizontal(cellI, cellJ);
                    result = 2;
                }
            }
            
        } else if (passType == 3) {
            if (state.getVerticalState(cellI, cellJ) == -1) {
                if (vertexLinks[cellI][cellJ].sameSet(vertexLinks[cellI+1][cellJ])) {
                    clearVertical(cellI, cellJ);
                    result = 2;
                } else if (cellJ > 0 && cellJ < width && cellLinks[cellI][cellJ-1].sameSet(cellLinks[cellI][cellJ])) {
                    if (cellLinks[cellI][cellJ-1].samePart(cellLinks[cellI][cellJ]))
                        clearVertical(cellI, cellJ);
                    else
                        fillVertical(cellI, cellJ);
                    result = 2;
                }
            }
            
        } else if (passType == -1) {
            if (puzzle.getClue(cellI, cellJ) == 3) {
                if (cellJ < width-1 && puzzle.getClue(cellI, cellJ+1) == 3) {
                    fillVertical(cellI, cellJ);
                    fillVertical(cellI, cellJ+1);
                    fillVertical(cellI, cellJ+2);
                    if (cellI > 0) clearVertical(cellI-1, cellJ+1);
                    if (cellI < height-1) clearVertical(cellI+1, cellJ+1);
                    result = 2;
                }
                if (cellI < height-1 && puzzle.getClue(cellI+1, cellJ) == 3) {
                    fillHorizontal(cellI, cellJ);
                    fillHorizontal(cellI+1, cellJ);
                    fillHorizontal(cellI+2, cellJ);
                    if (cellJ > 0) clearHorizontal(cellI+1, cellJ-1);
                    if (cellJ < width-1) clearHorizontal(cellI+1, cellJ+1);
                    result = 2;
                }
            }
        }
        
        return result;
    }
    
    private void clearHorizontal(int i, int j) {
        state.setHorizontalState(i, j, 0);
        data[i][j+1] &= ~MASK_DOWN;
        data[i+1][j+1] &= ~MASK_UP;
        if (i > 0 && i < height) cellLinks[i-1][j].union(cellLinks[i][j], true);
    }
    
    private void clearVertical(int i, int j) {
        state.setVerticalState(i, j, 0);
        data[i+1][j] &= ~MASK_RIGHT;
        data[i+1][j+1] &= ~MASK_LEFT;
        if (j > 0 && j < width) cellLinks[i][j-1].union(cellLinks[i][j], true);
    }
    
    private void fillHorizontal(int i, int j) {
        state.setHorizontalState(i, j, 1);
        data[i][j+1] &= MASK_DOWN;
        data[i+1][j+1] &= MASK_UP;
        vertexLinks[i][j].union(vertexLinks[i][j+1]);
        if (i > 0 && i < height) cellLinks[i-1][j].union(cellLinks[i][j], false);
    }
    
    private void fillVertical(int i, int j) {
        state.setVerticalState(i, j, 1);
        data[i+1][j] &= MASK_RIGHT;
        data[i+1][j+1] &= MASK_LEFT;
        vertexLinks[i][j].union(vertexLinks[i+1][j]);
        if (j > 0 && j < width) cellLinks[i][j-1].union(cellLinks[i][j], false);
    }
    
    public boolean nextStep() {
        boolean result = false;
        
        switch (passType) {
            case -1:
                if (cellJ < width-1) {
                    cellJ++;
                } else {
                    if (cellI < height-1) {
                        cellJ = 0;
                        cellI++;
                    } else {
                        cellI = -1;
                        cellJ = -1;
                        passType = 0;
                    }
                }
                break;
            case 0:
                if (cellJ < width) {
                    cellJ++;
                } else {
                    if (cellI < height) {
                        cellJ = -1;
                        cellI++;
                    } else {
                        cellI = 0;
                        cellJ = 0;
                        passType = 1;
                    }
                }
                break;
            case 1:
                if (cellJ < width-1) {
                    cellJ++;
                } else {
                    if (cellI < height) {
                        cellJ = 0;
                        cellI++;
                    } else {
                        cellI = height;
                        cellJ = width;
                        passType = 2;
                    }
                }
                break;
            case 2:
                if (cellI > -1) {
                    cellI--;
                } else {
                    if (cellJ > -1) {
                        cellI = height;
                        cellJ--;
                    } else {
                        cellI = 0;
                        cellJ = 0;
                        passType = 3;
                    }
                }
                break;
            case 3:
                if (cellI < height-1) {
                    cellI++;
                } else {
                    if (cellJ < width) {
                        cellI = 0;
                        cellJ++;
                    } else {
                        cellI = -1;
                        cellJ = -1;
                        passType = 0;
                        result = true;
                    }
                }
                break;
        }
        return result;
    }
    
}
