// Copyright 2009, 2010, 2012 Christopher Smith
// 
// This file is part of smithers.util.
// 
// smithers.util is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
// 
// smithers.util is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.util. If not, see <http://www.gnu.org/licenses/>.

package smithers.util;

import java.util.HashMap;
import java.util.Map;

/**
 * Class to parse command-line arguments. This is intended as a Java
 * replacement to POSIX's getopt() with several extensions. The valid options
 * are specified in a {@code String} parameter defined below. The valid options
 * may be altered at any time during parsing using the methods {@code
 * setOptions()} and {@code addOptions()}. This class allows both short options
 * (such as {@code -x}) and long options (such as {@code --ecks}). In both
 * cases an option may take no argument or an optional or required
 * option-argument. The author apologises for the length of some of the
 * following sentences, but is making no effort to improve them.
 * <br>
 * A program argument with a single leading hyphen ({@code '-'}) consists of
 * one or more short options. If an option takes an option-argument, then the
 * remainder of the argument, if non-empty, is the option-argument. If the
 * remainder of the argument is empty and an option-argument is required, then
 * the next argument is the option-argument, otherwise there is no
 * option-argument.
 * <br>
 * A program argument with two leading hyphens is a long option. If the
 * argument contains an equals sign ({@code '='}), then the portion of the
 * argument between the second hyphen and the equals sign (exclusive) is the
 * option, which must take an (optional or required) option-argument, and the
 * remainder of the argument after the equals sign is the option-argument. If
 * the argument does not contain an equals sign, the whole argument after the
 * second hyphen is the option and there is no option-argument unless the
 * option requires one, in which case the next argument is the option-argument.
 * <br>
 * An argument which which begins with a non-hyphen is taken as an operand, as
 * is an argument which is literally "-". If an argument is equal to {@code
 * "--"}, it is neither an option nor an operand, but all arguments after it
 * are interpreted as operands, even if they have leading hyphens.
 * <h4>Option strings</h4>
 * The option string passed to this class (any parameter with the name {@code
 * optionString} is interpreted as a sequence of <em>option
 * specifications</em>, optionally separated by spaces. The space separation is
 * needed in some cases to prevent ambiguity (most commonly, after a long
 * option which takes no arguments). An <em>option specification</em> consists
 * of an <em>option name specification</em> followed by an <em>option-argument
 * specification</em>. The <em>option-argument specification</em> consists of a
 * colon {@code ':'} to indicate a required argument, a question-mark {@code
 * '?'} for an optional argument or else is empty if no option-argument is
 * allowed. The <em>option name specification</em> is either a <em>short option
 * name specification</em> or a <em>long option name specification</em>. A
 * <em>short option name specification</em> consists of a single character
 * (other than {@code ' '} or {@code '-'}, which is the name of the option. A
 * <em>long option name specification</em> consists of a hyphen, then
 * optionally another hyphen and a single character (the <em>short equivalent
 * name</em>), then the name of the option - a sequence of one or more
 * characters other than space, colon or question-mark. A long option with a
 * <em>short equivalent name</em> is reported to the application as if it where
 * the corresponding short option. Note that the short equivalent option need
 * not be a valid short option - in this case when this class reports such a
 * short option the application knows that it must actually have received the
 * long option.
 */
public class GetOpt {
    
    private class LongOptSpec {
        char shortOpt = '-';
        int argumentSpec = ARGUMENT_ILLEGAL;
    }
    
    private static final int ARGUMENT_ILLEGAL  = 0,
                             ARGUMENT_OPTIONAL = 1,
                             ARGUMENT_REQUIRED = 2;
    
    /** The last option read. */
    public String option = null;
    
    /** The argument to the option, or {@code null}. */
    public String optionArg = null;
    
    /** The index of the next program argument to parse. */
    public int optionIndex = 0;
    
    private Map<Character, Integer> shortOpts = new HashMap<Character, Integer>();
    private Map<String, LongOptSpec> longOpts = new HashMap<String, LongOptSpec>();
    
    private int optionPos = 0;
    
    /**
     * Constructs a new GetOpt object which accepts no arguments.
     */
    public GetOpt() {
    }
    
    /**
     * Constructs a new GetOpt object with the given option string. Using this
     * constructor is equivalent to using the no-argument constructor and then
     * calling {@code addOptions(optionString)} on the newly constructed
     * object.
     * 
     * @param optionString the valid options, as described in the class summary
     */
    public GetOpt(String optionString) {
        addOptions(optionString);
        
//         // Testing.
//         for (Map.Entry<Character, Integer> e : shortOpts.entrySet()) {
//             switch (e.getValue()) {
//                 case ARGUMENT_ILLEGAL:
//                     System.out.println("Short option: -" + e.getKey());
//                     break;
//                 case ARGUMENT_OPTIONAL:
//                     System.out.println("Short option: -" + e.getKey() + "[arg]");
//                     break;
//                 case ARGUMENT_REQUIRED:
//                     System.out.println("Short option: -" + e.getKey() + " arg");
//                     break;
//             }
//         }
//         for (Map.Entry<String, LongOptSpec> e : longOpts.entrySet()) {
//             switch (e.getValue().argumentSpec) {
//                 case ARGUMENT_ILLEGAL:
//                     System.out.println("Long option: --" + e.getKey() + " (getOpt returns '" + e.getValue().shortOpt + "')");
//                     break;
//                 case ARGUMENT_OPTIONAL:
//                     System.out.println("Long option: --" + e.getKey() + "[=arg] (getOpt returns '" + e.getValue().shortOpt + "')");
//                     break;
//                 case ARGUMENT_REQUIRED:
//                     System.out.println("Long option: --" + e.getKey() + "[=]arg (getOpt returns '" + e.getValue().shortOpt + "')");
//                     break;
//             }
//         }
    }
    
    private void readOptionString(String optionString, boolean add)
    {
        int i = 0;
        while (i < optionString.length()) {
            if (optionString.charAt(i) == ' ') {
                i++;
            } else if (optionString.charAt(i) == '-') {
                // Long option.
                if (++i >= optionString.length()) throw new IllegalArgumentException("Invalid long option specification.");
                
                LongOptSpec spec = new LongOptSpec();
                if (optionString.charAt(i) == '-') {
                    if (++i >= optionString.length()) throw new IllegalArgumentException("Invalid long option specification.");
                    spec.shortOpt = optionString.charAt(i);
                    ++i;
                }
                int j;
jloop:          for (j = i ; j < optionString.length(); j++) {
                    switch (optionString.charAt(j)) {
                        case ' ':
                            break jloop;
                        case '?':
                            spec.argumentSpec = ARGUMENT_OPTIONAL;
                            break jloop;
                        case ':':
                            spec.argumentSpec = ARGUMENT_REQUIRED;
                            break jloop;
                    }
                }
                if (j == i) throw new IllegalArgumentException("Invalid long option specification.");
                
                if (add)
                    longOpts.put(optionString.substring(i, j), spec);
                else
                    longOpts.remove(optionString.substring(i, j));
                i = j + 1;
            } else {
                // Short option.
                char opt = optionString.charAt(i);
                int argSpec = ARGUMENT_ILLEGAL;
                if (++i < optionString.length()) {
                    switch (optionString.charAt(i)) {
                        case '?':
                            argSpec = ARGUMENT_OPTIONAL;
                            i++;
                            break;
                        case ':':
                            argSpec = ARGUMENT_REQUIRED;
                            i++;
                            break;
                    }
                }
                if (add)
                    shortOpts.put(opt, argSpec);
                else
                    shortOpts.remove(opt);
            }
        }
    }
    
    /**
     * Sets the options that this object allows.
     * 
     * @param optionString the valid options, as described in the class summary
     */
    public void setOptions(String optionString) {
        shortOpts.clear();
        longOpts.clear();
        readOptionString(optionString, true);
    }
    
    /**
     * Add new options to those that can be read by this object.
     * 
     * @param optionString the valid options, as described in the class summary
     */
    public void addOptions(String optionString) {
        readOptionString(optionString, true);
    }
    
    /**
     * Remove some of the options that can be read by this object. The
     * option-argument specifications need not match those originally supplied,
     * the same applies to the short-equivalent names of long options. It is
     * also not an error if the option to be removed does not exist. Thus
     * calling {@code setOptions("ab?c:d?h--hhelp")} followed by {@code
     * removeOptions("bcd:-help?")} will result in the same state as a calling
     * {@code setOptions("ah")}.
     * 
     * @param optionString the options to remove
     */
    public void removeOptions(String optionString) {
        readOptionString(optionString, false);
    }
    
    private int getShortOpt(String[] args) {
        char opt = args[optionIndex].charAt(optionPos++);
        option = Character.toString(opt);
        if (!shortOpts.containsKey(opt)) {
            // Invalid option.
            opt = '?';
            if (optionPos == args[optionIndex].length()) {
                optionIndex++;
                optionPos = 0;
            }
        } else {
            switch (shortOpts.get(opt)) {
                case ARGUMENT_ILLEGAL:
                    if (optionPos == args[optionIndex].length()) {
                        optionIndex++;
                        optionPos = 0;
                    }
                    break;
                case ARGUMENT_OPTIONAL:
                    if (optionPos < args[optionIndex].length()) optionArg = args[optionIndex].substring(optionPos);
                    optionIndex++;
                    optionPos = 0;
                    break;
                case ARGUMENT_REQUIRED:
                    if (optionPos < args[optionIndex].length()) {
                        optionArg = args[optionIndex].substring(optionPos);
                    } else if (optionIndex + 1 < args.length) {
                        optionArg = args[++optionIndex];
                    } else {
                        opt = ':';
                    }
                    optionIndex++;
                    optionPos = 0;
                    break;
            }
        }
        return opt;
    }
    
    /**
     * Parse the next option from the arguments.
     * <!-- TODO: describe the effect on members. -->
     * <br>
     * If a short option was read, returns the option name. If a long option
     * was read, returns the short equivalent name, if any, or else {@code
     * '-'}. If an option was read which but an option-argument was required
     * but missing or illegal but given, returns {@code ':'}. If an invalid
     * option was read, returns {@code '?'}. If the end of options was reached,
     * returns {@code -1}.
     * 
     * @param args the program arguments, as passed to {@code Main()}
     * @return the short option read, or one of {@code '-',':','?',-1} as
     *         described above
     * @throws RuntimeException if an error occurs due to the application
     *         altering {@code optionIndex} or supplying a different {@code
     *         args} parameter during the parsing of a group of short options
     */
    public int getOpt(String[] args) {
        option = optionArg = null;
        
        if (optionIndex >= args.length) {
            // Passed the end of the arguments.
            optionIndex = args.length;
            optionPos = 0;
            return -1;
        } else if (optionPos > 0) {
            // Already reading short options from args[optionIndex].
            if (optionPos >= args[optionIndex].length()) {
                // This shouldn't happen unless someone is messing with args or optionIndex between calls.
                // May want to do something other than this.
                throw new RuntimeException("Oy!");
            }
            return getShortOpt(args);
        } else if (args[optionIndex].length() <= 1 || args[optionIndex].charAt(0) != '-') {
            // Operand, not option.
            return -1;
        } else if (args[optionIndex].charAt(1) != '-') {
            // One or more short options.
            optionPos = 1;
            return getShortOpt(args);
        } else if (args[optionIndex].length() == 2) {
            // "--".
            optionIndex++;
            return -1;
        } else {
            // Long option.
            int i = args[optionIndex].indexOf('=');
            if (i < 0) {
                option = args[optionIndex].substring(2);
            } else {
                option = args[optionIndex].substring(2, i);
                optionArg = args[optionIndex].substring(i + 1);
            }
            optionIndex++;
            LongOptSpec spec = longOpts.get(option);
            if (spec == null) {
                // Invalid long option.
                return '?';
            } else {
                switch (spec.argumentSpec) {
                    case ARGUMENT_ILLEGAL:
                        if (optionArg != null) return ':';
                        break;
                    case ARGUMENT_OPTIONAL:
                        break;
                    case ARGUMENT_REQUIRED:
                        if (optionArg == null) {
                            if (optionIndex >= args.length) return ':';
                            optionArg = args[optionIndex++];
                        }
                        break;
                }
                return spec.shortOpt;
            }
        }
    }
    
}
